%
% Implementation of Exposure Fusion
%
% written by Tom Mertens, Hasselt University, August 2007
% e-mail: tom.mertens@gmail.com
%
% This work is described in
%   "Exposure Fusion"
%   Tom Mertens, Jan Kautz and Frank Van Reeth
%   In Proceedings of Pacific Graphics 2007
%
%
% Usage:
%   result = exposure_fusion(I,m);
%   Arguments:
%     'I': represents a stack of N color images (at double
%       precision). Dimensions are (height x width x 3 x N).
%     'm': 3-tuple that controls the per-pixel measures. The elements 
%     control contrast, saturation and well-exposedness, respectively.
%
% Example:
%   'figure; imshow(exposure_fusion(I, [0 0 1]);'
%   This displays the fusion of the images in 'I' using only the well-exposedness
%   measure
%
% =========================================================================
% WARNING!
%
% This function has been modified so as to handle the case were the
% sequence contains only gray images.
% Furthermore, it accepts supplementary parameters:
% muE, sigE, Nlev, WeightMaps. 
% Each of them is optional.
%
% Usage:
%   [R, W] = exposure_fusion_gray (I, m, muE, sigE, Nlev, WeightMaps)
% 
% For example, to reduce the number of scales, use
% [R, W] = exposure_fusion(I, m, [], [], Nlev)
%
% This function was copy-pasted and adapted from the one in directory
% ~/thesis/Articles/04_ExposureFusion/experiments/170510_EFSI/EF/ ...
% exposure_fusion.m
% 
% Charles Hessel, CMLA, ENS Paris-Saclay
% Created on Tuesday, April 10, 2018.

function [R, W] = exposure_fusion_gray (I, m, muE, sigE, Nlev, WeightMaps)

r = size(I,1);
c = size(I,2);
d = size(I,3);
N = size(I,4);
graySeq = ( d == 1 );

if ~exist('muE','var')
    muE = .5;
elseif isempty(muE)
    muE = .5;
end

if ~exist('sigE','var')
    sigE = .2;
elseif isempty(sigE)
    sigE = .2;
end

if ~exist('Nlev','var')
    Nlev = 0;
elseif isempty(Nlev)
    Nlev = 0;
end

if exist('WeightMaps','var')
    W = WeightMaps;
else
    W = ones(r,c,N);

    %compute the measures and combines them into a weight map
    contrast_parm = m(1);
    sat_parm = m(2);
    wexp_parm = m(3);

    if (contrast_parm > 0)
        if ~graySeq
            W = W.*contrast(I).^contrast_parm;
        else
            W = W.*contrast_gray(I).^contrast_parm;
        end
    end
    if (sat_parm > 0)
        if ~graySeq
            W = W.*saturation(I).^sat_parm;
        else
            error('Trying to measure Saturation of gray image. Set saturation weigth to 0.');
        end
    end
    if (wexp_parm > 0)
        if ~graySeq
            W = W.*well_exposedness(I,muE,sigE).^wexp_parm;
        else
            W = W.*well_exposedness_gray(I,muE,sigE).^wexp_parm;
        end
    end

    if ~exist('Nlev','var')
        Nlev = 0;
    end

    %normalize weights: make sure that weights sum to one for each pixel
    W = W + 1e-12; %avoids division by zero
    W = W./repmat(sum(W,3),[1 1 N]);
end

% limit number of levels in the pyramid ("0" means no limit)
if Nlev == 0
    nlev_max = floor(log(min(r,c)) / log(2)); % Charles
else
    nlev_max = Nlev;
end

% create empty pyramid
pyr = gaussian_pyramid(zeros(r,c,d),nlev_max); % Modif Charles (option nlev=3)
nlev = length(pyr);

% multiresolution blending
for i = 1:N
    % construct pyramid from each input image
	pyrW = gaussian_pyramid(W(:,:,i),nlev_max);
	pyrI = laplacian_pyramid(I(:,:,:,i),nlev_max);
    
    % blend
    for l = 1:nlev
        if ~graySeq
            w = repmat(pyrW{l},[1 1 3]);
        else
            w = pyrW{l};
        end
        pyr{l} = pyr{l} + w.*pyrI{l};
    end
end

% reconstruct
R = reconstruct_laplacian_pyramid(pyr);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% contrast measure
function C = contrast(I)
h = [0 1 0; 1 -4 1; 0 1 0]; % laplacian filter
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    mono = rgb2gray(I(:,:,:,i));
    C(:,:,i) = abs(imfilter(mono,h,'replicate'));
end
function C = contrast_gray(I)
h = [0 1 0; 1 -4 1; 0 1 0]; % laplacian filter
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    C(:,:,i) = abs(imfilter(I(:,:,:,i),h,'replicate'));
end

% saturation measure
function C = saturation(I)
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    % saturation is computed as the standard deviation of the color channels
    R = I(:,:,1,i);
    G = I(:,:,2,i);
    B = I(:,:,3,i);
    mu = (R + G + B)/3;
    C(:,:,i) = sqrt(((R - mu).^2 + (G - mu).^2 + (B - mu).^2)/3);
end

% well-exposedness measure
function C = well_exposedness(I, muE, sigE)
sig = sigE;
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    R = exp(-.5*(I(:,:,1,i) - muE).^2/sig.^2);
    G = exp(-.5*(I(:,:,2,i) - muE).^2/sig.^2);
    B = exp(-.5*(I(:,:,3,i) - muE).^2/sig.^2);
    C(:,:,i) = R.*G.*B;
end
function C = well_exposedness_gray(I, muE, sigE)
sig = sigE;
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    C(:,:,i) = exp(-.5*(I(:,:,:,i) - muE).^2/sig.^2);
end
