function simulatedExposureFusion ( ...
    alpha, nImgs, beta, softClip, lambda, labCS, ...
    Wc, Ws, We, ...
    SCB, blackSat, whiteSat)
% Simulated Exposure Fusion.
% This function takes a color image as input, and improves the "clarity" by
% increasing the contrast of the luminance layer.
% It works as follows:
%   1) compute the luminance layer
%   2) simulate a bracketed exposure sequence from it
%   3) merge the sequence using exposure fusion
%   4) add color back
%
% All parameters are strings (IPOL Demo system constraint).
% They are converted in the function, as follows.
%   alpha       double > 1          maximum contrast
%   nImgs       integer > 1         number of images
%   beta        double \in (0,1]    reduced dynamic range
%   softClip    bool                soft/hard clipping functions
%   lambda      double              speed of derivative annulation in softC
%   labCS       bool                use Lab Color Space
%   Wc          double              exposure fusion: contrast
%   Ws          double              exposure fusion: saturation
%   We          double              exposure fusion: well-exposedness
%   SCB         bool                apply simplest color balance
%   blackSat    double < 50         SCB: dark saturation percentage
%   whiteSat    double < 50         SCB: bright saturation percentage
%
% Charles Hessel, CMLA, ENS Paris-Saclay. Monday April 9, 2018.

%%% PREVIOUSLY
% Friday, April 6, 2018. CMLA.
% Copy of the script EEF_SmartNSmoothSeq.m from ~/thesis/Articles/04_EF.../
% I modify it here in order to clarify it.
% Final objective is to use it in the "Simulated Exposure Fusion" Workshop
% on IPOL (update of the previous one).

%%% Convert input parameters
alpha   = str2double(alpha);
nImgs   = str2double(nImgs);
beta    = str2double(beta);
softClip= logical(str2double(softClip));
lambda  = str2double(lambda);
labCS   = logical(str2double(labCS));
Wc      = str2double(Wc);
Ws      = str2double(Ws);
We      = str2double(We);
SCB     = logical(str2double(SCB));
blackSat= str2double(blackSat);
whiteSat= str2double(whiteSat);

%%% Check parameters
if alpha == 1, error('parameter alpha must be >1'); end
if beta < 0, error('parameter beta must be >0'); end
if beta > 1, error('parameter beta must be <1'); end
if Ws ~= 0
    warning(['parameter Ws was set to 0. ' ...
             'Only luminance is processed in SEF.']);
    Ws = 0;
end
if We ~= 0
    warning('parameter We should preferably be 0.');
end
if blackSat > 50, blackSat = 50; end
if whiteSat > 50, whiteSat = 50; end


%%% Fixed parameters
% muE     = .5;
% sigE    = .2;
% Nlev    = 0; % that means no limit for the number of scales
PRINT   = true;
OUTDIR  = '.';
FMT     = 'png';


%%% Load input image
Icol = imread('input_0.png');
Icol = im2double(Icol);


%%% Normalize the input
fprintf(['\nNormalisation of the input image with 0.01%% of\n' ...
    'clipped values in black and in white parts.\n']);
Icol = simplestColorBalance( Icol, .0001, .0001 );


%%% === apply filter to luminance only
if labCS
    L = rgb2lab(Icol); % convert input to L*a*b color space
    I = min(1,max(0,L(:,:,1)/100)); % L is the luminance
else
    I = rgb2gray(Icol);
    C = Icol./repmat(I+double(I==0),[1 1 3]);
end


%%% === Generete de sequence from luminance
fprintf('\nGenerating the input sequence... '); tic
if softClip
    [Seq, x, y, nBrigthImgs, nDarkImgs] = ...
        simulateSmartAndSmoothSeq (I, alpha, nImgs, beta, lambda);
else
    [Seq, x, y, nBrigthImgs, nDarkImgs] = ...
        simulateSmartSeq (I, alpha, nImgs, beta);
end
fprintf('Done. Elapsed time: %.3f seconds.\n',toc);


%%% === Exposure Fusion
fprintf('\nApplying exposure fusion... '); tic
% [R, W] = exposure_fusion(Seq, [Wc Ws We], muE, sigE, Nlev);
[R, W] = exposure_fusion_gray(Seq, [Wc Ws We]);
fprintf('Done. Elapsed time: %.3f seconds.\n',toc);


%%% === Simplest Color Balance

if logical(SCB)
    fprintf('\nNormalizing the output image.\n');
    R = simplestColorBalance(R, whiteSat/100, blackSat/100);
else
    fprintf('\nClipping out-of-range values (if necessary).\n');
    R = min(1, max(0, R)); % <-- clip out-of-range values
end


%%% === Restore color
fprintf('\nRestoring color.\n');
if labCS
    Rcol = lab2rgb(cat(3,R*100,L(:,:,2:3)));
else
    Rcol = C.*repmat(R,[1 1 3]);
end


%%% Measure the quality of the input according to TMQI
fprintf('\nComputing TMQI between input and output image:\n');
hdrImage = Icol*(2^16 -1);
ldrImage = Rcol*(2^8  -1);
[Q, S, N, ~, ~] = TMQI(hdrImage, ldrImage);
%%% Print result to stdout
fprintf(['Q = %.3f (TMQI score)\n' ...
         '--> S = %.3f (structural fidelity score)\n' ...
         '--> N = %.3f (statistical naturalness score)\n'], Q, S, N);


%%% Compute dispersion map
%%% Use the luminance
fprintf('\nComputing the dispersion map.\n');
dMap = dispersionMap(max(0, min(1, I)), max(0, min(1, R)));
if PRINT
    imwrite(uint8(255*dMap/max(dMap(:))), parula(256), ...
        sprintf('%s/dispersionMap.png',OUTDIR));
end


%%% Print sequence and associated weights
if PRINT
    for k = 1:nImgs+1
        imwrite(uint8(Seq(:,:,:,k)*255), ...
            sprintf('%s/sequ_n%d.%s',OUTDIR,k-1,FMT));
        imwrite(uint8(W(:,:,k)*255), ...
            sprintf('%s/sequ_n%d_weights.%s',OUTDIR,k-1,FMT));
    end
end


%%% Print output color fused image
if PRINT
    imwrite(uint8(255*Rcol), sprintf('%s/output.%s', OUTDIR, FMT));
end


%%% Print remapping functions
if PRINT
    fh = figure('visible','off');
    plot(x,y,'LineWidth',1.25); axis([0 1 0 1]); axis image
    leg = cell(0);
    for k=-nBrigthImgs:nDarkImgs
        leg{nBrigthImgs+1+k} = ['k=' num2str(k)];
    end
    legend(leg,'FontSize',14,'Location','SouthEast');
    set(gca,'position',[0 0 1 1],'units','normalized')
    set(gcf,'PaperUnits','Inches','PaperPosition',[0 0 5.12 5.12])
    print('-dpng',sprintf('%s/remappingFunctions.png',OUTDIR),'-r100');
end

