function [Seq, x, y, nBrigthImgs, nDarkImgs] = simulateSmartAndSmoothSeq ...
    (Y, alpha, nImgs, beta, lambda)
% simulateSmartAndSmoothSeq generates a sequence of images from the
% provided input image. The sequence simulates a exposure bracketing.
%
% See also function simulateSmartSeq.
% 
% Y     : input image, gray level, in [0 1]
% alpha : maximal contrast factor
% nImgs : number of images to generate
% beta  : non-saturated dynamic range of generated images
% lambda: speed of contrast attenuation for the "smooth clipping"
% Seq   : output sequence
% x,y   : remapping functions used for the images in the sequence
%
% Charles Hessel, CMLA, ENS Paris-Saclay
% Created on Friday, April 6, 2018.

%%% --> Initialization
Seq = [];
x   = (0:.001:1)';
y   = [];

%%% --> Measure median value
%%% median value of input image
%%% I used to use the lab luminance for that measure (even if the rest of
%%% the algorithm used the rgb2gray function).
Imed = median(Y(:));

%%% --> Compute the number of bright and dark images from Imed and other
%%%     parameters
nBrigthImgs = round((1-Imed)*nImgs); % (1-Imed) is a percentage
nDarkImgs = nImgs - nBrigthImgs; % remaining images are for bright parts

%%% --> Prepare the remapping functions 
%%% f_dark   improves contrast in bright areas (saturation at 0)
%%% f_bright improves contrast in dark areas (saturation at 1)
nImgsMax = max(nBrigthImgs,nDarkImgs);
f_dark   = @(t,k) alpha^(k/nImgsMax) * (t-1)+1 ;
f_bright = @(t,k) alpha^(k/nImgsMax) * t ;

%%% This function gives the clipping center
%%% Warning, here in this function k varies in [-nImgs/2,+nImgs/2]
cc   = @(k)( k*(1-beta)/nImgs + .5 );

%%% Function for the "smooth clipping"
gam = @(alpha) beta/(2*alpha*(lambda-1));
fun = @(x,alpha)(   (abs(x) <= beta/2/alpha).*(alpha*x) + ...
                    (abs(x) >  beta/2/alpha).*(...
                    sign(x).*( alpha*gam(alpha)*...
                    log( (abs(x+double(x==0)) - gam(alpha)*(lambda-2) )*...
                    exp(lambda-1)/gam(alpha) ) ) ) ...
                    );
clip = @(t,k) max(0,min(1, fun( t - cc(k), 1) + cc(k) ));

%%% First set of images improves contrast in dark areas
for k = nBrigthImgs:-1:1
    Seq = cat(4,Seq, f_bright(Y,k) );
    y   = cat(2,y  , f_bright(x,k) ); % draw remapping functions
end

%%% In the middle, use original image
Seq = cat(4,Seq,Y);
y   = cat(2,y  ,x); % draw remapping functions

%%% Second set of images improves contrast in bright areas
for k = 1:nDarkImgs
    Seq = cat(4,Seq, f_dark(Y,k) );
    y   = cat(2,y  , f_dark(x,k) ); % draw remapping functions
end

%%% Clipping of Sequence (assume Images are sorted from the "brightest" to
%%% the "darkest"); Clip the unmodified input too.
for k = 1:nImgs+1
    n = k - 1 - nImgs/2;
    Seq(:,:,:,k) = clip(Seq(:,:,:,k),n);
    y(:,k)       = clip(y(:,k)      ,n);
end

