function [Seq, x, y, nBrigthImgs, nDarkImgs] = simulateSmartSeq ...
    (I, alpha, nImgs, beta)
% simulateSmartSeq generates a sequence of images from the
% provided input image. The sequence simulates a exposure bracketing.
%
% See also function simulateSmartAndSmoothSeq.
% 
% I     : input image, gray level, in [0 1]
% alpha : maximal contrast factor
% nImgs : number of images to generate
% beta  : non-saturated dynamic range of generated images
% lambda: speed of contrast attenuation for the "smooth clipping"
% Seq   : output sequence
% x,y   : remapping functions used for the images in the sequence
%
% Charles Hessel, CMLA, ENS Paris-Saclay
% Created on Friday, April 6, 2018.

Seq = [];
x   = (0:.001:1)';
y   = [];


%%% Computes number of bright and dark images in function of median value
Imed = median(I(:)); % median value of input image (in [0 1])
nBrigthImgs = round((1-Imed)*nImgs); % (1-Imed) is a percentage
nDarkImgs = nImgs - nBrigthImgs; % remaining images are for bright parts


%%% f_dark   improves contrast in bright areas (saturation at 0)
%%% f_bright improves contrast in dark areas (saturation at 1)
nImgsMax = max(nBrigthImgs,nDarkImgs);
f_dark   = @(t,k) max(0, alpha^(k/nImgsMax) * (t-1)+1 );
f_bright = @(t,k) min(1, alpha^(k/nImgsMax) * t );

%%% Warning, here in this function k varies in [-nImgs/2,+nImgs/2]
cc   = @(k)( k*(1-beta)/nImgs + .5 ); % clipping center
clip = @(t,k) max( cc(k) - beta/2, min( cc(k) + beta/2, t ));

%%% First set of images improves contrast in dark areas
for k = nBrigthImgs:-1:1
    Seq = cat(4,Seq, f_bright(I,k) );
    y   = cat(2,y  , f_bright(x,k) ); % draw remapping functions
end

%%% In the middle, use original image
Seq = cat(4,Seq,I);
y   = cat(2,y  ,x); % draw remapping functions

%%% Second set of images improves contrast in bright areas
for k = 1:nDarkImgs
    Seq = cat(4,Seq, f_dark(I,k) );
    y   = cat(2,y  , f_dark(x,k) ); % draw remapping functions
end

%%% Clipping of Sequence (assume Images are sorted from the "brightest" to
%%% the "darkest"); Clip the unmodified input too.
for k = 1:nImgs+1
    n = k - 1 - nImgs/2;
    Seq(:,:,:,k) = clip(Seq(:,:,:,k),n);
    y(:,k)       = clip(y(:,k)      ,n);
end